/* ----------------------------------------------------------------------------
 *
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 *
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* perf08_db.h was generated from the DDL file perf08_db.sdl. */
#include "perf08_db.h"

/* perf08_db_dbd.h was generated from the DDL file perf08_db.ddl. using ddlp with the -c option */
#include "perf08_db_dbd.h"

/*! Number of levels in the BOM */
#define NUM_LEVELS 6

/*! Number of members per level in the BOM */
#define NUM_MEMBERS 8

/*! Context structure for the Perf08 example */
typedef struct _PERF08_CTX
{
    int32_t         current_level;   /*!< Level currently being processed */
    int32_t         max_level;       /*!< Maximum BOM level */
    int32_t         max_members;     /*!< Maximum number of members per BOM level */
    int32_t         count;           /*!< Count of BOM members */
    uint32_t        key_state_size;  /*!< Size of the internal RDM keystate tables */
    TFS_HANDLE      hTFS;            /*!< RDM TFS handle */
    DB_TASK        *task;            /*!< RDM task */
} PERF08_CTX;

/*! \file perf08.c
 * \brief Source code for the RDM Perf08 example
 */
 
/*!
 * \mainpage Perf08
 * \author Jeffrey R. Parsons, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf08
 *
 * The Perf08 example creates and reads a Bill-of-Materials using relational
 * index definitions.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * A new Bill-of-Materials is created
 *   * The Bill-of-Materials is read
 *   * Database is cleaned up
 *
 * \section Perf08 Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf08_db.ddl. The example utilizes 
 * the RDM Core DDL syntax and specifies a database suitable for storing a
 * Bill-of-Materials.
 *
 * \include perf08_db.ddl
 *
 * The schema was compiled using the RDM ddlp utility with the -c option to
 * generate an embedded database dictionary (used by the RDM d_open_ptr API).
 *
 * \code ddlp -c -d perf08_db_db.ddl \endcode
 *
 * The database dictionary information is embedded inside the application so 
 * an external DBD file stored on the file system is not required.
 *
 * \section RDM Functions Used
 * The following RDM functions are illustrated:
 *   * d_tfsinitEx/d_tfsterm
 *   * d_opentask/d_closetask
 *   * d_open_ptr/d_close
 *   * d_initialize
 *   * d_keyszstate
 *   * d_fillnew
 *   * d_keyrdstate/d_keywrstate
 *   * d_keyfind
 *   * d_recread
 *   * d_keyfrst
 *   * d_crread
 *   * d_keynext
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Initialize the RDM runtime library for use in the perf08 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use 
 * the STANDALONE transactional TFS implementation.  It also creates
 * an RDM runtime task and opens the "perf08_db" database in exclusive ('x') mode.
 * Exclusve mode does not require database locks or transactions.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t initialize(
	PERF08_CTX  *ctx)    /**< [in/out] Pointer to the application context */
{
    int32_t  rdm_status = S_OKAY;
    uint64_t start_time;
    uint64_t end_time;    

    /* Initialize the TFS to use the single-user configuration (STANDALONE) */    
    start_time = psp_timeNanoSecs();
    rdm_status = d_tfsinitEx(NULL, NULL, &ctx->hTFS, TFS_TYPE_STANDALONE);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error initializing the TFS (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Create the RDM DB_TASK */
    rdm_status = d_opentask(&ctx->task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening the RDM task (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Open the perf08_db database in exclusive mode (does not require locks or tranactions) */
    rdm_status = d_open_ptr("perf08_db", "x", &perf08_db_dbd, sizeof(perf08_db_dbd), ctx->task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening the perf08_db database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Initialize the database (creates new database files truncating any existing data) */    
    rdm_status = d_initialize(ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error initializing the perf08_db database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Get the size of the buffer required to save the databases keystate tables */
    rdm_status = d_keyszstate(RBOM, &ctx->key_state_size, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error read key state size: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    end_time = psp_timeNanoSecs();

	ctx->max_level = NUM_LEVELS;
	ctx->max_members = NUM_MEMBERS;
    
    /* Display information */
    printf("\n");
    printf("Initializing with STANDALONE configuration\n");
    display_elapsed_time("Preparing a new database:\t\t\t", start_time, end_time);
    printf("\tDisk-based\n");
    printf("\tOpened exclusively\n");

    return rdm_status;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM runtime task
 * used in the perf08 example.  It also terminates the STANDALONE TFS used
 * in the example. 
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
void cleanup(
    PERF08_CTX *ctx)   /**< [in] Pointer to the application context */
{
    uint64_t start_time;
    uint64_t end_time;

    start_time = psp_timeNanoSecs();
    if(ctx->task)
    {
        /* Close any open databases */
        d_close(ctx->task);

        /* Close the task */
        d_closetask(ctx->task);
    }

    /* Terminate the TFS */
    d_tfsterm(ctx->hTFS);

    end_time = psp_timeNanoSecs();

    printf("\n");
    display_elapsed_time("Cleanup:\t\t\t\t\t", start_time, end_time);
}

/*! \brief Generate a 15-character alpha-numeric part id
 *
 * This function generates a part id for an item 
 *
 */
void bom_random_id(
    PERF08_CTX *ctx,        /**< [in] Pointer to the application context */
    char       *string)     /**< [out] 15-character alpha part id */
{
    sprintf(string, "%-15.15d", ++(ctx->count));
}

/*! \brief Recursive rountine to build one level of a bill by adding components to a parent
 *
 * This function recursively builds a full bill of materials in the Core08
 * database.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t build_bill(
    const char *parent,    /**< [in] The item_id of the parent part in the bill */
    PERF08_CTX *ctx)       /**< [in] Pointer to the application context */
{
    int32_t      rdm_status;
    int32_t      ii;
    char         id_code[SIZEOF_RID_CODE];
    struct rbill bill_rec;
    struct ritem item_rec;

    ctx->current_level++;

    for (ii=0; ii<ctx->max_members; ii++) 
    {
        /* create an RITEM record which will be a component of the parent
           RITEM record whose id code was passed into this function
        */
        item_rec.rcost = 1.0L;
        bom_random_id(ctx, (char *)(item_rec.rid_code));
        if (ctx->current_level < ctx->max_level)
        {
            item_rec.rcomponent_count = ctx->max_members;
        }
        else
        {
            /* A leaf node */
            item_rec.rcomponent_count = 0;
        }

        /* new component ITEM record */
        if ( (rdm_status = d_fillnew(RITEM, &item_rec, ctx->task, CURR_DB)) != S_OKAY)
        {
            fprintf(stderr, "Error creating ITEM record: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        }

        /* create an BILL record, with relational references to the parent
           and component ITEM records - this BILL record provides a
           linkage in a many-to-many relationship
        */
        strcpy((char *)bill_rec.rparent, (char *)parent);
        strcpy((char *)bill_rec.rcomponent, (char *)item_rec.rid_code);
        bill_rec.rquantity = 1.0L;
        bill_rec.rsequence = ii;
        bill_rec.rlevel = ctx->current_level;
        rdm_status = d_fillnew(RBILL, &bill_rec, ctx->task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error creating BILL record: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }

        /* if we are not at the bottom of the tree, attach further
           components to this component ITEM, by recursively calling
           this function
        */
        if (ctx->current_level < ctx->max_level) 
        {
            strcpy((char *)id_code, (char *)item_rec.rid_code);
            rdm_status = build_bill(id_code, ctx);
        }
    }

    ctx->current_level--;
    return rdm_status;
}

/*! \brief Top level function to create a BOM
 *
 * This function recursively builds a full bill of materials in the Core07
 * database.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t create_bom(
    PERF08_CTX *ctx) /**< [in] Pointer to the application context */
{
    int32_t      rdm_status;
    struct ritem item_rec;
    char         id_code[SIZEOF_RID_CODE];
    uint64_t start_time;
    uint64_t end_time;

    start_time = psp_timeNanoSecs();

    /* create top level ITEM record */
    bom_random_id(ctx, id_code);
    strcpy((char *)item_rec.rid_code,  (char *)id_code);
    item_rec.rcomponent_count = ctx->max_members;
    item_rec.rcost = 1.0;
    if ((rdm_status = d_fillnew(RITEM, &item_rec, ctx->task, CURR_DB)) != S_OKAY) 
    {
		fprintf(stderr, "Error creating ITEM record: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* recursive call to build multi-level bill */
    rdm_status = build_bill(id_code, ctx);

    end_time = psp_timeNanoSecs();

    display_elapsed_time("\t\t", start_time, end_time);
    return rdm_status;
}

/*! \brief Recursive rountine roll up cost from lower levels of a BOM
 *
 * This function recursively rolls-up the cost of a BOM.  The costs are
 * only stored at the lowest levels of the BOM.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */

int32_t get_cost(
    const char *parent,      /**< [in]  The item id of the current component we are summing */
    double     *total_cost,  /**< [out] Pointer to area to put the total cost */
    PERF08_CTX *ctx)         /**< [in]  Pointer to the application context */
{
    int32_t            rdm_status;
    double             dCost;
    double             dTotalCost;
    int32_t            component_count;
    struct rbom        bom;
    struct rbill       bill_local;
    struct ritem       item_rec;
    void              *key_state;

    key_state = psp_cGetMemory(ctx->key_state_size, TAG0);
    if(!key_state)
    {
        return S_NOMEMORY;
    }
    
    /* get the current key state for the RBOM key - on exit, this function
       repositions to this state, so that calls to d_keynext in the previous
       level of recursion will find the right value
    */
    rdm_status = d_keyrdstate(RBOM, key_state, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        return rdm_status;
    }

    /* use RID_CODE key to find the parent item */
    rdm_status = d_keyfind(RID_CODE, parent, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        return rdm_status;
    }
    
    /* read the whole record */
    rdm_status = d_recread(&item_rec, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        return rdm_status;
    }

    component_count = item_rec.rcomponent_count;
    if (component_count == 0) 
    {
        /* no components for this parent - return the cost of this
           bottom-level item
        */
        *total_cost = item_rec.rcost;
        return rdm_status;
    }

    /* there is at least one component, so go down a level - use RBOM key
       to find the first RBILL record that refers to the parent item
    */
    strcpy((char *)bom.rparent, (char *)parent);
    bom.rsequence = 0;
    rdm_status = d_keyfind(RBOM, &bom, ctx->task, CURR_DB);  /* find first bill record */
    if(rdm_status != S_OKAY)
    {
        return rdm_status;
    }
    dTotalCost = 0.0L;

    /* use RBOM key to go through all RBILL records that refer to the
       parent item
    */
    while (rdm_status == S_OKAY)
    {
        /* read bill rec to get component ID */
        rdm_status = d_recread(&bill_local, ctx->task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            return rdm_status;
        }

        /* this component may also be a parent - call this function
           recursively to get costs for the component tree attached
           to the RITEM record
        */
        rdm_status = get_cost((const char *)bill_local.rcomponent, &dCost, ctx);
        if(rdm_status == S_OKAY)
        {
            dTotalCost += (dCost * bill_local.rquantity); 
        }
        else
        {
            return rdm_status;
        }

        if(--component_count == 0) break;

        /* find next bill record */
        rdm_status = d_keynext(RBOM, ctx->task, CURR_DB);
    }

    /* reposition to start position in RBOM key */
    rdm_status = d_keywrstate(RBOM, key_state, ctx->task, CURR_DB);
    psp_freeMemory(key_state, TAG0);

    *total_cost = dTotalCost;
    return rdm_status;
}

/*! \brief Top level function for rolling the BOM
 *
 * This function recursively rolls up and calculate the full cost
 * of a BOM.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t rollup_bom(
    PERF08_CTX *ctx) /**< [in] Pointer to the application context */
{
    int32_t  rdm_status;
    char     id_code[SIZEOF_RID_CODE];
    double   total_cost = 0.0;
    uint64_t start_time;
    uint64_t end_time;

    start_time = psp_timeNanoSecs();

    /* Navigate the top level ITEM record */
    rdm_status = d_keyfrst(RID_CODE, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error navigating to first ITEM record: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    rdm_status = d_crread(RID_CODE, id_code, ctx->task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error reading first ITEM record: %s - %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    rdm_status = get_cost(id_code, &total_cost, ctx);
    end_time = psp_timeNanoSecs();
    
    printf("\tTotal cost is $%10.02lf", total_cost);
    display_elapsed_time("\t\t", start_time, end_time);
    
    return rdm_status;
}

/*! \brief Main function for perf08 example
 *
 * The function initializes the RDM environment and runs the insert operations.
 * over a number of transaction block sizes
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main()
{
    PERF08_CTX  ctx;
    int32_t     rdm_status;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf08"));

    memset(&ctx, 0, sizeof(PERF08_CTX));

    /* Initialize the TFS, task, and open/initialize the database */
    rdm_status = initialize(&ctx);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(&ctx);
        return 1;
    }

    /* Build the BOM */
    printf("\nCreating BOM with %d levels and %d members per level\n", ctx.max_level, ctx.max_members);
    printf("\tBuilding bill of materials");
    create_bom(&ctx);
    
    /* Read the BOM - rollup cost */
    printf("\nRolling up BOM cost\n");
    rollup_bom(&ctx);

    cleanup(&ctx);

    printf("\nHit <Enter> to continue . . .");
    getchar();

	return 0;
}
